/**
 *  =============================================================
 *  SymbolTable.java: Maintains a table for namespace scope
 * 
 *  Values = map from strings to stacks, i.e.,
 *
 *    [a]->[value1,0],[value2,0]
 *
 *  table = a stack of bindings
 *  Maintains a table for namespace scope
 *  =============================================================
 */

package symbol01;

import java.util.ArrayList;
import java.util.List;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;
import java.util.Map;
import java.util.Iterator;

public class SymbolTable {
   private int   depth;
   private Stack table;
   private Map  values;
   boolean DEBUG = true;
    
   public SymbolTable() {	
      depth = 0;
      table  = new Stack();
      values = new HashMap();
   }
  
   public void clear() {
      depth = 0; 
      table.clear();
      values.clear();
   }
  
   // Retrieve the current depth

   public int getDepth() {
      return depth;
   }

   public boolean withinElement() {
      return depth > 0;
   }

   // =============================================
   // Adds a name/value pair
   // =============================================

   public void put (String name, String value) {

      if ( DEBUG == true ) {
         System.out.printf("--- In SymbolTable.put(): Add [name, value ] = [%s, %s]\n", name, value );
      }

      table.push( new Symbol(name,value,depth) );

      if ( values.containsKey(name) == false ) {
         Stack valueStack = new Stack();
         valueStack.push(value);
         values.put(name,valueStack);
      } else {
         Stack valueStack = (Stack) values.get(name);
         valueStack.push(value);
      }
   }
    
   // =============================================
   // Gets the value for a variable
   // =============================================

   public String get (String name) {
      Stack valueStack = (Stack) values.get(name);
      if (valueStack==null || valueStack.isEmpty()) 
         return null;
      return (String) valueStack.peek();
   }

   public Set getAll(String name) {
      HashSet result = new HashSet();
      Iterator i = table.iterator();
      while (i.hasNext()) {
        Symbol s = (Symbol) i.next();
        if (name.equals(s.getName()))
            result.add(s.getValue());
      }
      return result;
   }

   // =============================================
   // add a new highest level scope to the table
   // =============================================

   public void openScope() {

      if (DEBUG == true) {
         System.out.printf("--- In SymbolTable.openScope(): depth = %2d\n", depth + 1 );
      }

      depth++;
   }

   // =============================================
   // Remove the highest level scope from the table
   // =============================================
  
   public void closeScope() {

      if (DEBUG == true) {
         System.out.printf("--- In SymbolTable.closeScope(): depth = %2d\n", depth );
      }

      // Get the top binding

      Symbol symbol = (Symbol) table.peek();
      int symbolDepth = symbol.depth;

      // Check if it needs to be popped of the table

      while (symbolDepth == depth && !table.isEmpty()) {
         symbol = (Symbol) table.pop();

         // pop its value as well

         Stack valueStack = (Stack) values.get(symbol.name);
         valueStack.pop();

         // check the next binding

         if (!table.isEmpty()) {
            symbol = (Symbol) table.peek();
            symbolDepth = symbol.depth;
         } else break;
      }
      depth--;
   }

   // =============================================
   // String value of symbol table ....
   // =============================================

   public String toString() {
      StringBuffer buf = new StringBuffer();
      buf.append( String.format("\n") );
      buf.append( String.format("Symbol Table: depth = %2d\n", depth ) );
      buf.append( String.format("============================================\n") );

      Iterator i = table.iterator();
      while(i.hasNext()) {
         Symbol symbol = (Symbol) i.next();
         buf.append( String.format("%s\n", symbol ) );
      }

      buf.append( String.format("============================================\n") );
      buf.append( String.format("\n") );
      return buf.toString();
  }

  // ============================================================
  // Exercise methods in symbol table ...
  // ============================================================

  public static void main(String args[]) throws Exception {

    System.out.printf("--- Test symbol table stack            ... \n");
    System.out.printf("--- ================================== ... \n");

    // Create empty symbol table ....

    SymbolTable st = new SymbolTable();

    // Open scope for level 1 ...

    System.out.printf("--- Open scope for level 1 ... \n");

    st.openScope();
    st.put("x","horizontal 1");
    st.put("y","vertical 1");
    st.put("z","out of page 1");

    System.out.printf("--- Retrieve x: %s ...\n", st.get("x") );
    System.out.printf("--- Retrieve y: %s ...\n", st.get("y") );

    // Open scope for level 2 ...

    System.out.printf("--- Open scope for level 2 ... \n");

    st.openScope();
    st.put("x","horizontal 2");
    st.put("y","vertical 2");

    // Open scope for level 3 ...

    System.out.printf("--- Open scope for level 3 ... \n");

    st.openScope();
    st.put("x","horizontal 3");

    // Open scope for level 4 ...

    System.out.printf("--- Open scope for level 4 ... \n");

    st.openScope();
    st.put("x","horizontal 4");

    System.out.printf("--- Retrieve x: %s ...\n", st.get("x") );

    // Print symbol table having depth 4 ...

    System.out.printf("--- Print symbol table having depth 4 ... \n");

    System.out.print(st);

    // Retrieve symbols from depth 4 ...

    System.out.printf("--- Retrieve symbols from level 4 ... \n");

    System.out.printf("--- Retrieve x: %s ...\n", st.get("x") );
    System.out.printf("--- Retrieve y: %s ...\n", st.get("y") );
    System.out.printf("--- Retrieve z: %s ...\n", st.get("z") );

    // Close out levels 4 and 3 ...

    System.out.printf("--- Closeout levels 4 and 3 ... \n");

    st.closeScope();
    st.closeScope();

    // Print symbol table having depth 2 ...

    System.out.printf("--- Print symbol table having depth 2 ... \n");

    System.out.print(st);

    // Retrieve symbols from depths 1 and 2 ...

    System.out.printf("--- Retrieve x: %s ...\n", st.get("x") );
    System.out.printf("--- Retrieve y: %s ...\n", st.get("y") );
    System.out.printf("--- Retrieve z: %s ...\n", st.get("z") );

    // Close out level 2 ...

    System.out.printf("--- Closeout level 2 ... \n");

    st.closeScope();

    // Print symbol table having depth 1 ...

    System.out.printf("--- Print symbol table having depth 1 ... \n");

    System.out.print(st);

    // Close out level 1 ...

    System.out.printf("--- Closeout level 1 ... \n");

    st.closeScope();

    System.out.print(st);

    System.out.printf("--- ================================== ... \n");
    System.out.printf("--- Finished !! ... \n");

  }
}
