/**
 * ==================================================================
 * SymbolTableStack.java: Implementation of the symbol table stack.
 *
 * Copyright (c) 2009 by Ronald Mak
 * For instructional purposes only.  No warranties.
 * ==================================================================
 */

package symbol02;

import java.util.ArrayList;
import symbol02.*;

public class SymbolTableStackImpl extends ArrayList<SymbolTable>
       implements SymbolTableStack {

    private int currentNestingLevel;    // current scope nesting level
    private SymbolTableEntry programId; // entry for the main program id

    // Constructor methods ...

    public SymbolTableStackImpl() {
        this.currentNestingLevel = 0;
        add(SymbolTableFactory.createSymbolTable(currentNestingLevel));
    }

    /**
     * Setter.
     * @param entry the symbol table entry for the main program identifier.
     */
    public void setProgramId(SymbolTableEntry id)
    {
        this.programId = id;
    }

    /**
     * Getter.
     * @return the symbol table entry for the main program identifier.
     */
    public SymbolTableEntry getProgramId()
    {
        return programId;
    }

    /**
     * Getter.
     * @return the current nesting level.
     */
    public int getCurrentNestingLevel()
    {
        return currentNestingLevel;
    }

    /**
     * Return the local symbol table which is at the top of the stack.
     * @return the local symbol table.
     */
    public SymbolTable getLocalSymbolTable()
    {
        return get(currentNestingLevel);
    }

    /**
     * Push a new symbol table onto the symbol table stack.
     * @return the pushed symbol table.
     */
    public SymbolTable push()
    {
        SymbolTable symTab = SymbolTableFactory.createSymbolTable(++currentNestingLevel);
        add(symTab);

        return symTab;
    }

    /**
     * Push a symbol table onto the symbol table stack.
     * @return the pushed symbol table.
     */

    public SymbolTable push(SymbolTable symTab) {
        ++currentNestingLevel;
        add(symTab);
        return symTab;
    }

    /**
     * Pop a symbol table off the symbol table stack.
     * @return the popped symbol table.
     */

    public SymbolTable pop() {
        SymbolTable symTab = get(currentNestingLevel);
        remove(currentNestingLevel--);
        return symTab;
    }

    /**
     * Create and enter a new entry into the local symbol table.
     * @param name the name of the entry.
     * @return the new entry.
     */

    public SymbolTableEntry enterLocal(String name) {
        return get(currentNestingLevel).enter(name);
    }

    /**
     * Look up an existing symbol table entry in the local symbol table.
     * @param name the name of the entry.
     * @return the entry, or null if it does not exist.
     */

    public SymbolTableEntry lookupLocal(String name) {
        return get(currentNestingLevel).lookup(name);
    }

    /**
     * Look up an existing symbol table entry throughout the stack.
     * @param name the name of the entry.
     * @return the entry, or null if it does not exist.
     */

    public SymbolTableEntry lookup(String name) {
        SymbolTableEntry foundEntry = null;

        // Search the current and enclosing scopes.
        for (int i = currentNestingLevel; (i >= 0) && (foundEntry == null); --i) {
            foundEntry = get(i).lookup(name);
        }

        return foundEntry;
    }
}
